"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
const abstract_leveldown_1 = require("abstract-leveldown");
const level_mount_iterator_1 = require("./level-mount-iterator");
const easier_abstract_leveldown_1 = require("easier-abstract-leveldown");
const concat_1 = require("./util/concat");
/**
 * LevelMount, like file system directory mounts and somewhat like the opposite of subleveldown.
 * Can be used *with* subleveldown. Examples:
 *
 * const hybrid_db = levelup(levelmount({
 *   db: leveldown('root'),
 *   options: {}, // options for leveldown
 *   mounts: [
 *     {
 *       mount: '!tmp!',
 *       db: memdown(),
 *     },
 *     {
 *       mount: '!var!',
 *       db: someotherdown(),
 *       options: {}, // someotherdown options for _open
 *     }
 *   ],
 * })
 *
 * hybrid_db.put('hello', 'world') // ends up in root leveldown with key `hello`
 * hybrid_db.put('!tmp!blah', 'bleh') // ends up in memdown with key `blah`
 *
 * const var_db = subleveldown(hybrid_db, 'var', { separator: '!' })
 * var_db.put('goodbye', 'world') // ends up in someotherdown with key `goodbye`
 */
class LevelDOWNMount extends abstract_leveldown_1.AbstractLevelDOWN {
    constructor(opts) {
        super(undefined);
        this._mounts = [];
        // Sort mount locations by length (largest first)
        this._mounts = [
            {
                db: opts.db,
                mount: '',
                options: opts.options,
            },
            ...(opts.mounts || []),
        ].sort((a, b) => b.mount.length - a.mount.length);
    }
    // Resolve the mount and internal path that a given key falls under
    _resolve_mount(key) {
        // TODO: improve mount search complexity?
        // TODO: handle keys that aren't strings?
        // Coerce key to string
        const key_str = String(key);
        // Select the first mount which matches the prefix of the key
        //  (the root mount with prefix '' will get matched as a fallback)
        for (const { db, mount } of this._mounts) {
            if (key_str.slice(0, mount.length) === mount) {
                return {
                    db,
                    mount,
                    path: key_str.slice(mount.length),
                };
            }
        }
        throw new Error('An impossible error occured');
    }
    _open(opts, callback) {
        Promise.all(this._mounts.map(({ db, options }) => new Promise((resolve, reject) => {
            db.open(options, (err) => {
                if (err)
                    reject(err);
                else
                    resolve();
            });
        }))).then(() => process.nextTick(callback)).catch((err) => callback(err));
    }
    _close(callback) {
        Promise.all(this._mounts.map(({ db }) => new Promise((resolve, reject) => {
            db.close((err) => {
                if (err)
                    reject(err);
                else
                    resolve();
            });
        }))).then(() => process.nextTick(callback)).catch((err) => callback(err));
    }
    _get(key, options, callback) {
        // Forward to underlying store
        const { db, path } = this._resolve_mount(key);
        db.get(path, options, callback);
    }
    _put(key, val, options, callback) {
        // Forward to underlying store
        const { db, path } = this._resolve_mount(key);
        db.put(path, val, options, callback);
    }
    _del(key, options, callback) {
        // Forward to underlying store
        const { db, path } = this._resolve_mount(key);
        db.del(path, options, callback);
    }
    _batch(array, options, cb) {
        // Group batch operations by resolved mount
        const dbs = {};
        const ops = {};
        for (const op of array) {
            const { mount, db, path } = this._resolve_mount(op.key);
            if (ops[mount] === undefined) {
                ops[mount] = [];
                dbs[mount] = db;
            }
            ops[mount].push(Object.assign({}, op, { key: path }));
        }
        // Forward batch options to underlying stores
        Promise.all(Object.keys(ops).map((mount) => new Promise((resolve, reject) => {
            dbs[mount].batch(ops[mount], options, (err) => {
                if (err)
                    reject(err);
                else
                    resolve();
            });
        }))).then(() => process.nextTick(cb)).catch((err) => process.nextTick(cb, err));
    }
    _iterator(options) {
        return new level_mount_iterator_1.LevelDOWNMountIterator(this, options);
    }
    // If the underlying store supports the EasierAbstractLevelDOWN changes
    //  extension, we'll capture those changes and emit them on this store.
    changes() {
        const newEmitter = new easier_abstract_leveldown_1.EasierLevelDOWNEmitter();
        for (const { mount, db } of this._mounts) {
            if (db.changes !== undefined) {
                db.changes().onPut((key, value) => newEmitter.emitPut(concat_1.concat(mount, key), value)).onDel((key) => newEmitter.emitDel(concat_1.concat(mount, key))).onBatch((array) => newEmitter.emitBatch(array.map((op) => {
                    if (op.type === 'put') {
                        return {
                            type: op.type,
                            key: concat_1.concat(mount, op.key),
                            value: op.value,
                        };
                    }
                    else if (op.type === 'del') {
                        return {
                            type: op.type,
                            key: concat_1.concat(mount, op.key),
                        };
                    }
                    else
                        throw new Error(`Unrecognized batch operation '${op.type}'`);
                })));
            }
        }
        return newEmitter;
    }
}
exports.LevelDOWNMount = LevelDOWNMount;
//# sourceMappingURL=level-mount.js.map